package com.sleepycat.je.cleaner;

import com.sleepycat.je.DatabaseException;
import com.sleepycat.je.config.EnvironmentParams;
import com.sleepycat.je.dbi.DbConfigManager;
import com.sleepycat.je.dbi.EnvConfigObserver;
import com.sleepycat.je.dbi.EnvironmentImpl;
import com.sleepycat.je.utilint.DaemonRunner;

public class CleanerDaemon implements DaemonRunner, EnvConfigObserver {
	private FileProcessorDaemon[] threads;
	private Cleaner cleaner;
	private String name;
	private EnvironmentImpl env;
	int nDeadlockRetries;

	public CleanerDaemon(EnvironmentImpl env, String name, Cleaner cleaner) throws DatabaseException {
		this.name=name;
		this.cleaner = cleaner;
		this.env=env;
		threads = new FileProcessorDaemon[0];

		/* Initialize mutable properties and register for notifications. */
		envConfigUpdate(env.getConfigManager());
		env.addConfigObserver(this);
	}

	/**
	 * Process notifications of mutable property changes.
	 */
	public void envConfigUpdate(DbConfigManager cm) throws DatabaseException {
		nDeadlockRetries = cm.getInt(EnvironmentParams.CLEANER_DEADLOCK_RETRY);

		
		int nThreads = cm.getInt(EnvironmentParams.CLEANER_THREADS);
		assert nThreads > 0;

		if (nThreads != threads.length) {

			/* Shutdown threads when reducing their number. */
			for (int i = nThreads; i < threads.length; i += 1) {
				if (threads[i] != null) {
					threads[i].shutdown();
					threads[i] = null;
				}
			}

			/* Copy existing threads that are still used. */
			FileProcessorDaemon[] newThreads = new FileProcessorDaemon[nThreads];
			for (int i = 0; i < nThreads && i < threads.length; i += 1) {
				newThreads[i] = threads[i];
			}

			/* Don't lose track of new threads if an exception occurs. */
			threads = newThreads;

			/* Start new threads when increasing their number. */
			for (int i = 0; i < nThreads; i += 1) {
				if (threads[i] == null) {
					FileProcessor fileProcessor=new FileProcessor(name + '-' + (i + 1), env,
							cleaner, cleaner.profile, cleaner.fileSelector);
					threads[i] = new FileProcessorDaemon(name + '-' + (i + 1), env,
							this, fileProcessor);
				}
			}
		}

	}
	
	
	/*
	 * Delegate the run/pause/wakeup/shutdown DaemonRunner operations. We always
	 * check for null to account for the possibility of exceptions during thread
	 * creation. Cleaner daemon can't ever be run if No Locking mode is enabled.
	 */

	public void runOrPause(boolean run) {
		if (!env.isNoLocking()) {
			for (int i = 0; i < threads.length; i += 1) {
				if (threads[i] != null) {
					threads[i].runOrPause(run);
				}
			}
		}
	}

	public void wakeup() {
		for (int i = 0; i < threads.length; i += 1) {
			if (threads[i] != null) {
				threads[i].wakeup();
			}
		}
	}

	public void requestShutdown() {
		for (int i = 0; i < threads.length; i += 1) {
			if (threads[i] != null) {
				threads[i].requestShutdown();
			}
		}
	}

	public void shutdown() {
		for (int i = 0; i < threads.length; i += 1) {
			if (threads[i] != null) {
				threads[i].shutdown();
				threads[i].clearEnv();
				threads[i] = null;
			}
		}
	}

	public int getNWakeupRequests() {
		int count = 0;
		for (int i = 0; i < threads.length; i += 1) {
			if (threads[i] != null) {
				count += threads[i].getNWakeupRequests();
			}
		}
		return count;
	}

	public boolean areThreadsRunning() {
		for (int i = 0; i < threads.length; i += 1) {
			if (threads[i] != null) {
				return threads[i].isRunning();
			}
		}
		return false;
	}

}
